import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import javax.swing.JOptionPane; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class booby_foot extends PApplet {

/*
Software to load pictures of seabird feet, visually move nodes to relevant locations, and output node coordinates to csv

Written April 2018 by Kevin Anderson, supervised by David Anderson (WFU)
Modified December 2018 

Source code written in the Processing language (https://processing.org/), version 3.3.2
*/





File[] files;
Node[] nodes;
TextBox[] boxes;
Button[] radios;
                       //   Heel,    D1,   D2,   D3,   D4a,   D4b,   CalHeel,   CalD1,   CalD2,   CalD3,   CalD4a,   CalD4b,   CalA1,   CalA2,   CalA3
int[] defaultLx = new int[] {499,   316, 1057, 1262,  1040,  1133,       500,     322,    1057,    1258,     1040,     1040,     397,    1143,    1114};
int[] defaultLy = new int[] {619,   282,  279,  571,   826,   824,       651,     313,     304,     604,      855,      886,     281,     280,     914};
int[] defaultRx = new int[] {486,   551, 1066, 1312,  1181,  1273,       485,     552,    1063,    1314,     1183,     1181,     439,    1154,    1091};
int[] defaultRy = new int[] {676,   885,  913,  527,   208,   208,       652,     910,     940,     562,      264,      294,     889,     913,     239};


PImage currentImage;
int currentImageIndex;
Node activeNode;
Node pin;
TextBox activeBox;
Button inputButton;
Button outputButton;
Button nextButton;
Button quitButton;
Button skipButton;
String defaultOutputFile;
String screenshotFolder;

int sidePaneWidth = 155;
String footSide;

PVector shift;

boolean updateNextFrame = false;

ArrayList<String> data = new ArrayList<String>();

public void setup(){
  
  inputButton = new Button(new PVector(10,10), "Photo Folder");
  outputButton = new Button(new PVector(10, 50), "Save Data");
  nextButton = new Button(new PVector(10, height-80), "");
  skipButton = new Button(new PVector(10, height-120), "Skip Photo");
  quitButton = new Button(new PVector(10, height-40), "Quit");
  boxes = new TextBox[0];
  nodes = new Node[0];
  radios = new Button[0];
  shift = new PVector(0,0);
  updateNextFrame = true;
  footSide = "";
  defaultOutputFile = "";
}

public void drawTickLine(float startX, float startY, float stepX, float stepY, int n){
  int tickSize = 5;
  float x = startX;
  float y = startY;
  float endX = startX + n * stepX;
  float endY = startY + n * stepY;
  stroke(0,255,0);
  strokeWeight(5);
  line(startX, startY, endX, endY);
  for(int i = 0; i < n; i++){
    x += stepX;
    y += stepY;
    line(x-tickSize, y, x+tickSize, y);
    line(x, y-tickSize, x, y+tickSize);
  }
  strokeWeight(1);
}

public void update(){
  background(100);
  pushMatrix();
  translate(shift.x, shift.y);
  if(currentImage != null){
    image(currentImage, 0, 0);
  }
  for(Node node: nodes){
    node.display();
  }
  
  // draw calibration lines as a visual cue for user
  if(activeNode != null){
    if(activeNode.label.length() >= 4 && activeNode.label.substring(1,4).equals("Cal")){
      // get a handle on the CalHeel node
      Node calHeel = null;
      for(Node node: nodes){
        if(node.label.substring(1).equals("CalHeel")){
          calHeel = node;
          break;
        }
      }

      String label = activeNode.label.substring(4);

      for(TextBox box: boxes){
        if(box.label.equals(label+"X")){
          if(box.contents.length() > 0){
            int nx = parseInt(box.contents);
            if(nx != 0){
              float deltaX = activeNode.pos.x - calHeel.pos.x;
              float stepsize = deltaX / nx;
              drawTickLine(calHeel.pos.x, calHeel.pos.y, stepsize, 0, nx);
              drawTickLine(calHeel.pos.x, activeNode.pos.y, stepsize, 0, nx);
            }
          }
        } else if(box.label.equals(label+"Y")){
          if(box.contents.length() > 0){
            int ny = parseInt(box.contents);
            if(ny != 0){
              float deltaY = activeNode.pos.y - calHeel.pos.y;
              float stepsize = deltaY / ny;
              drawTickLine(calHeel.pos.x, calHeel.pos.y, 0, stepsize, ny);
              drawTickLine(activeNode.pos.x, calHeel.pos.y, 0, stepsize, ny);
            }
          }
        }
      }
    }
  }
  if(pin != null){
    pin.display();
  }
  popMatrix();
  fill(200);
  stroke(200);
  rect(0,0,sidePaneWidth,height);
  for(TextBox box: boxes){
    box.display();
  }
  for(int i = 0; i < radios.length; i++){
    radios[i].display();
  }
  inputButton.display();
  outputButton.display();
  nextButton.display();
  skipButton.display();
  quitButton.display();
  if(files != null && currentImageIndex < files.length && files[currentImageIndex] != null){
    text(files[currentImageIndex].getName(), 10, height - 140);
  }
}

public void reset(){
  
  char prefix = '?';
  String currentPath = files[currentImageIndex].getName();

  if(currentPath.indexOf('L') >= 0){
    prefix = 'l';
    footSide = "L";
  }
  if(currentPath.indexOf('R') >= 0){
    prefix = 'r';
    footSide = "R";
  }
  if(currentPath.indexOf('L') >= 0 && currentPath.indexOf('R') >= 0){
    prefix = '?';
    footSide = "?";
  }
  
  nodes = new Node[15];
  int[] x;
  int[] y;
  if(prefix == 'l'){
    x = defaultLx;
    y = defaultLy;
  } else if(prefix == 'r'){
    x = defaultRx;
    y = defaultRy;
  } else {
    // fill with garbage
    x = new int[nodes.length];
    y = new int[nodes.length];
    for(int i = 0; i < x.length; i++){
      x[i] = currentImage.width/2;
      y[i] = currentImage.height/2;
    }
  }
  String[] nodeLabels = new String[] {"Heel", "D1", "D2", "D3", "D4a", "D4b", 
                                  "CalHeel", "CalD1", "CalD2", "CalD3", "CalD4a", "CalD4b",
                                  "CalA1", "CalA2", "CalA3"};
  int green = color(0,255,0);
  int yellow = color(255,255,0);
  int pink = color(255, 105, 180);
  int[] colors = new int[] { green, green, green, green, green, green,
                                 yellow, yellow, yellow, yellow, yellow, yellow,
                                 pink, pink, pink};
  for(int i = 0; i < nodes.length; i++){
    nodes[i] = new Node(new PVector(x[i], y[i]), colors[i], prefix + nodeLabels[i]);
  }
  // CalHeel is more important
  nodes[6].textSize *= 2;
  nodes[6].selectedTextSize *= 2;

  activeNode = null;
  pin = new Node(new PVector(800, 800), color(255,165,0), "marker");
  // preserve value from last time
  String user = "";
  if(boxes != null && boxes.length > 0 && boxes[0] != null){
    user = boxes[0].contents;
  }
  boxes = new TextBox[19];
  boxes[0] = new TextBox(new PVector(10, 150), sidePaneWidth-20, "User");
  boxes[0].numeric = false;
  boxes[0].contents = user;
  String[] boxLabels = {"Heel", "D1", "D2", "D3", "D4a", "D4b", "A1", "A2", "A3"};
  int i = 1;
  for(String label: boxLabels){
    int boxY = 200 + (i/2) * 50;
    boxes[i] = new TextBox(new PVector(5, boxY), 70, label + "X");
    boxes[i+1]=new TextBox(new PVector(80,boxY), 70, label + "Y");
    i += 2;
  }
  
  radios = new Button[5];
  String[] radioLabels = {"0%", "<1%", "1-5%", "6-25%", ">25%"};
  for(i = 0; i < radios.length; i++){
    radios[i] = new Button(new PVector(5, 700 + i*40), i + ": " + radioLabels[i]);
    radios[i].buttonWidth = 75;
  }
  
  nextButton.label = "Move All Nodes";
  
  if(currentImage != null){
    shift = new PVector(width/2-currentImage.width/2,height/2-currentImage.height/2);
  } else {
    shift = new PVector(0,0);
  }
}

public boolean arrEQ(int[] a, int[] b){
  if(a.length != b.length){
    return false;
  }
  for(int i = 0; i < a.length; i++){
    if(a[i] != b[i]) return false;
  }
  return true;
}

public boolean allNodesMoved(){
  
  //if(true) return true;
  
  int[] defaultX;
  int[] defaultY;
  if(footSide.equals("L")){
    defaultX = defaultLx;
    defaultY = defaultLy;
  }
  else if(footSide.equals("R")){
    defaultX = defaultRx;
    defaultY = defaultRy;
  } else {
    defaultX = new int[nodes.length];
    defaultY = new int[nodes.length];
  }
  for(int i = 0; i < nodes.length; i++){
    if(nodes[i].pos.x == defaultX[i]) return false;
    if(nodes[i].pos.y == defaultY[i]) return false;
  }
  return true;
}

public void mousePressed(){
  // find which node was clicked, if any
  boolean found = false;
  for(Node node: nodes){
    if(node.hit(new PVector(mouseX-shift.x, mouseY-shift.y))){
      if(activeNode != null){
        activeNode.active = false;
      }
      if(activeBox != null){
        activeBox.active = false;
        activeBox = null;
      }
      activeNode = node;
      node.active = true;
      found = true;
      break;
    }
  }
  // push pin?
  if(!found && pin != null){
    if(pin.hit(new PVector(mouseX-shift.x, mouseY-shift.y))){
      if(activeNode != null){
        activeNode.active = false;
      }
      if(activeBox != null){
        activeBox.active = false;
        activeBox = null;
      }
      activeNode = pin;
      pin.active = true;
      found = true;
    }
  }
  // if not a node, maybe it was a box
  if(!found){
    for(TextBox box: boxes){
      if(box.hit(new PVector(mouseX, mouseY))){
        if(activeBox != null){
          activeBox.active = false;
        }
        // don't cancel a node activation when clicking in a box
        /*
        if(activeNode != null){
          activeNode.active = false;
          activeNode = null;
        }
        */
        activeBox = box;
        box.active = true;
        found = true;
        break;
      }
    }
  }
  // if not a node or box, maybe a button
  if(!found){
    PVector pos = new PVector(mouseX, mouseY);
    if(inputButton.hit(pos)){
      selectFolder("Select a folder to process:", "folderSelected");
    } else if(outputButton.hit(pos)){
      selectOutput("Choose where to save the data:", "outputSelected", new File(defaultOutputFile));
    } else if(nextButton.hit(pos)){
      if(allNodesMoved()){
        next(false);
      }
    } else if(skipButton.hit(pos)){
      next(true);
    } else if(quitButton.hit(pos)){
      int dialogResult = JOptionPane.showConfirmDialog (null, "Are you sure you want to quit?  Unsaved data will be lost","Warning",JOptionPane.NO_OPTION);
      if(dialogResult == JOptionPane.YES_OPTION){
        exit();
      }
    }
  }
  // maybe a radio button?
  if(!found){
    PVector pos = new PVector(mouseX, mouseY);
    for(int i = 0; i < radios.length; i++){
      if(radios[i].hit(pos)){
        for(int j = 0; j < radios.length; j++){
          radios[j].active = false;
        }
        radios[i].active = true;
      }
    }
  }
  // ok, user didn't click on anything
  if(!found){
    if(activeNode != null){
      activeNode.active = false;
      activeNode = null;
    }
    if(activeBox != null){
      activeBox.active = false;
      activeBox = null;
    }
  }
  updateNextFrame = true;
}

public void folderSelected(File selection) {
  if (selection != null) {
    String pictureDirectory = selection.getAbsolutePath();
    
    // save screenshot folder for later
    String datetime = getDatetime().replace(":", "");
    screenshotFolder = pictureDirectory + "\\screenshots_" + datetime;
    // get a list of sorted files, since listFiles doesn't guarantee it
    // also exclude directories
    File[] unsortedFiles = listFiles(pictureDirectory, "files");

    String[] unsortedFilepaths = new String[unsortedFiles.length];
    for(int i = 0; i < unsortedFiles.length; i++){
      unsortedFilepaths[i] = unsortedFiles[i].getAbsolutePath();
    }

    String[] sortedFilepaths = sort(unsortedFilepaths);
    files = new File[unsortedFiles.length];
    // O(n^2) sorting algo, ridic
    // can't be arsed to import hashmap for this
    // sure wish java could sort using subfields as keys, like python...
    for(int i = 0; i < files.length; i++){
      for(int j = 0; j < unsortedFiles.length; j++){
        if(unsortedFiles[j].getAbsolutePath().equals(sortedFilepaths[i])){
          files[i] = unsortedFiles[j];
          break;
        }
      }
    }
    
    String currentImagePath = files[0].getAbsolutePath();
    currentImage = loadImage(currentImagePath);
    currentImageIndex = 0;
    reset();
    updateNextFrame = true;
  }
}

public void outputSelected(File selection){
  if(selection != null){
    String outputFile = selection.getAbsolutePath();
    if(!outputFile.endsWith(".csv")){
      outputFile += ".csv";
    }
    try {
      PrintWriter output = createWriter(outputFile);
      String header = "File,Datetime,Spacer,Foot,";
      for(Node node: nodes){
        header += node.label + "X" + "," + node.label + "Y" + ",";
      }
      for(TextBox box: boxes){
        header += box.label + ",";
      }
      header += "Damage";
      output.println(header);
      
      for(String row: data){
        output.println(row);
      }
      // need to include the current photo's data
      output.println(getDataString());
      output.flush();
      output.close();
      defaultOutputFile = outputFile; 
    } catch(Exception e){
      JOptionPane.showMessageDialog(null, "Could not save output file!\nMaybe you have the file open in another window?\nYou can choose another filename and try again.\n" + e.toString());
    }
  }
}

public void saveScreenshot(){
  String outfile = screenshotFolder + "\\" + files[currentImageIndex].getName().replace(".JPG", "") + "_" + boxes[0].contents + ".png";
  try {
    save(outfile);
  } catch(Exception e){
      JOptionPane.showMessageDialog(null, "Could not save screenshot!\n" + outfile);
  }
}

public void next(boolean skipped){
  if(files == null){
    return;
  }
  
  if(currentImageIndex < files.length){
    // get data before incrementing index
    String row;
    if(skipped){
      row = getSpacerString();
    } else {
      row = getDataString();
    }
    data.add(row);
    
    // save a screenshot
    saveScreenshot();    
    
    currentImageIndex++;
  }
  
  if(currentImageIndex >= files.length){
    JOptionPane.showMessageDialog(null, "No more images to process!");
    PGraphics graphics = createGraphics(500,500);
    graphics.beginDraw();
    graphics.background(200);
    graphics.text("No more images!", 200,200);
    graphics.endDraw();
    currentImage = graphics.get();
    return;
  }
  currentImage = loadImage(files[currentImageIndex].getAbsolutePath());
  reset();
  updateNextFrame = true;
}

public void keyPressed(){
  if(key == ESC){
    key = 0;
    return;
  }
  String alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890";
  if(activeBox != null){
    if(key == BACKSPACE){
      if(activeBox.contents.length() > 0){
        activeBox.contents = activeBox.contents.substring(0, activeBox.contents.length()-1);
      }
    } else if((activeBox.numeric && (key >= '0' && key <= '9')) || 
              (!activeBox.numeric && (alphabet.indexOf(key) >= 0))){
      activeBox.contents += key;
    }
  }
  updateNextFrame = true;
}

public void mouseDragged(){
  
  if(activeNode != null && mouseX > sidePaneWidth){
    activeNode.pos = new PVector(constrain(mouseX-shift.x,0,currentImage.width), constrain(mouseY-shift.y,0,currentImage.height));
  } else {
    shift.x += mouseX - pmouseX;
    shift.y += mouseY - pmouseY;
  }
  if(allNodesMoved()){
    nextButton.label = "Next Photo";
  }
  updateNextFrame = true;
}

public String getSpacerString(){
  if(files == null) return "";
  String res = "";
  res += files[currentImageIndex].getName() + ",";
  res += getDatetime() + ",";
  res += "1";
  return res;
}

public String getDataString(){
  if(files == null) return "";
  String res = "";
  res += files[currentImageIndex].getName() + ",";
  res += getDatetime() + ",";
  res += "0" + ",";
  res += footSide + ",";
  for(Node node: nodes){
    res += node.pos.x + "," + node.pos.y + ",";
  }
  for(TextBox box: boxes){
    res += box.contents + ",";
  }
  for(int i = 0; i < radios.length; i++){
    if(radios[i].active){
      res += i;
      break;
    }
  }
  return res;
}

public String zpad(int n){
  if(n < 10){
    return "0" + n;
  }
  return "" + n;
}

public String getDatetime(){
  return year() + "-" + zpad(month()) + "-" + zpad(day()) + " " + 
         zpad(hour()) + ":" + zpad(minute()) + ":" + zpad(second());
}

public void draw(){
  if(updateNextFrame){
    update();
    updateNextFrame = false;
  }
}
class Button
{
  PVector pos;
  String label;
  int buttonWidth;
  int buttonHeight;
  boolean active; // only used for radio buttons
  
  Button(PVector pos, String label){
    this.pos = pos;
    this.label = label;
    buttonWidth = 100;
    buttonHeight = 30;
  }
  
  public void display(){
    fill(255);
    if(active){
      fill(255, 255, 0);
    }
    rect(pos.x, pos.y, buttonWidth, buttonHeight);

    stroke(0);
    fill(0);
    textSize(12);
    text(label, pos.x + 5, pos.y + buttonHeight - 8);
  }
  
  public boolean hit(PVector click){
    return click.x > pos.x && 
           click.x < pos.x + buttonWidth &&
           click.y > pos.y &&
           click.y < pos.y + buttonHeight;
  }
  
}
class Node
{
  PVector pos;
  float displayRadius;
  float clickRadius;
  int col;
  String label;
  boolean active;
  
  int textSize;
  int selectedTextSize;
  
  Node(PVector pos, int col, String label){
    this.pos = pos;
    this.col = col;
    this.label = label;
    this.displayRadius = 5;
    this.clickRadius = 20;
    this.active = false;
    this.textSize = 32;
    this.selectedTextSize = 40;
  }
  
  public void display(){
    fill(col);
    stroke(0);
    float rad = displayRadius;
    textSize(textSize);
    if(active){
      rad = rad * 1.2f;
      textSize(selectedTextSize);
    }
    ellipse(pos.x, pos.y, rad, rad);
    /*
    // black background behind text
    fill(0);
    for(int x = -2; x <= 2; x++){
      for(int y = -2; y <= 2; y++){
        text(label, pos.x + rad * 0.8 + x, pos.y + rad * 0.8 + y);
      }
    }
    */
    fill(col);
    text(label, pos.x + rad * 0.8f + 5, pos.y + rad * 0.8f);
  }
  
  public boolean hit(PVector click){
    return PVector.sub(click, pos).mag() < clickRadius;
  }
  
  
}
class TextBox
{
  PVector pos;
  String label;
  String contents;
  int boxWidth;
  int boxHeight;
  boolean active;
  boolean numeric;
  
  TextBox(PVector pos, int boxWidth, String label){
    this.pos = pos;
    this.label = label;
    this.contents = "";
    this.boxWidth = boxWidth;
    this.boxHeight = 25;
    this.numeric = true;
  }
  
  public void display(){
    fill(255);
    stroke(0);
    if(active){
      fill(255, 255, 0);
    }
    rect(pos.x, pos.y, boxWidth, boxHeight);

    stroke(0);
    fill(0);
    textSize(12);
    text(label + ":", pos.x + 5, pos.y + boxHeight - 8);
    text(contents, pos.x + textWidth(label + ":") + 10, pos.y + boxHeight - 8);


  }
  
  public boolean hit(PVector click){
    return click.x > pos.x && 
           click.x < pos.x + boxWidth &&
           click.y > pos.y &&
           click.y < pos.y + boxHeight;
  }
  
}
  public void settings() {  fullScreen(2); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "booby_foot" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
